/*
 * Soubor:  libsort.c
 * Datum:   31.12.2010 12:44
 * Autor:   Marek Nožka, nozka <@t> spseol <d.t> cz
 * Licence: GNU/GPL 
 * Úloha:   modul řadících  funkcí
 * Popis:   implementuje jednoduché řadící algoritmy Buble sort, Insert sort, 
 *          Select sort a Quick sort.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "libsort.h"
/////////////////////////////////////////////////////

/* *
 * Funkce inicializuje generátor náhodných čísel.
 * Konstanta SAMEARRAY rozhoduje, zda bude pole vždy stejné
 * nebo při každém spuštění jiné.
 */
void initgen(void)              //%.h
{
    if (SAMEARRAY == 1) {
        srand(SEED);
    } else {
        srand((unsigned int) time(NULL));
    }
}


/* 
 * funkce generuje pole array náhodných čísel 
 */
void generate(int start, int stop, int array[]) //%.h 
{
    for (int i = start; i <= stop; i++) {
        if (SMALLNUMBERS == 1)
            array[i] = rand() % 44;
        else
            array[i] = rand();
    }
}

/* 
 * Funkce vypíše pole na obrazovku.
 * Tato funkce je určena pouze pro ladění. 
 */
void show(int start, int stop, int array[])     //%.h
{
    for (int i = 0; i <= stop; i++) {
        printf("%d ", array[i]);
    }
    putchar('\n');
}

/* *
 * Funkce kontroluje, zda je pole správně seřazeno 
 */
bool control(int start, int stop, int array[])  //%.h
{
    for (int i = start; i <= stop - 1; i++) {
        if (array[i] > array[i + 1]) {
            return false;
        }
    }
    return true;
}

/* 
 * Funkce seřadí array metodou Bubble sort 
 * Bubble sort prochází pole znovu a znovu a přehodí vždy dva 
 * sousední prvky, které jsou v nesprávném pořadí.
 */
void bubbleSort(int start, int stop, int array[])       //%.h
{
    for (int j = start; j <= stop - 1; j++) {
        for (int i = start; i <= stop - 1; i++) {
            if (array[i] > array[i + 1]) {
                int tmp = array[i];
                array[i] = array[i + 1];
                array[i + 1] = tmp;
            }
        }
    }
}


/* 
 * Funkce seřadí array metodou Selection sort 
 * Selection sort buduje na začátku pole seřazenou oblast. Najde vždy 
 * nejmenší prvek v neseřazené oblasti a zařadí (vymění) těsně za konec 
 * seřazené oblasti.
 *          /------>
 * 2 6 7 | 29 23 18 11 17
 *         <--------^^
 * 2 6 7 11 | 23 18 29 17
 */
void selectSort(int start, int stop, int array[])       //%.h
{
    for (int i = start; i <= stop - 1; i++) {
        /* najde nejmenší */
        int min = i;
        for (int j = i + 1; j <= stop; j++) {
            if (array[j] < array[min]) {
                min = j;
            }
        }
        /* zařadí nejmenší za konec seřazené oblasti */
        int tmp = array[i];
        array[i] = array[min];
        array[min] = tmp;
    }
}

/* 
 * Funkce seřadí array metodou Insertion sort 
 * Insert Sort prochází prvky postupně a každý další nesetříděný 
 * prvek zařadí na správné místo do již setříděné posloupnosti tj. 
 * najde pro prvek pozici, na kterou podle své velikosti patří. 
 * Prvky o větší hodnotě posune doprava.
 * 2 7 23 29 | 18 6 11 17
 *       <-----^^  
 * 2 7 18 23 29 | 6 11 17
 *         -->                      
 */
void insertSort(int start, int stop, int array[])       //%.h
{
    for (int i = start; i <= stop; i++) {
        int value = array[i];   //prvek, pro který hledám pozici
        int j = i - 1;          // zde končí již setříděná oblast
        /* hledám pozici a zároveň prvky posouvám doprava */
        while (j >= 0 && array[j] > value) {
            array[j + 1] = array[j];
            j = j - 1;
        }
        array[j + 1] = value;   // zařadím prvek
    }

}

/* 
 * Funkce seřadí array metodou Quick sort -- rozdělením 
 * 
 */
void quickSort(int start, int stop, int array[])        //%.h
{
    int pivot = array[(stop + start) / 2];
    int left = start;
    int right = stop;
    while (left < right) {
        /* hledám zleva prvky větší než pivot */
        while (array[left] < pivot) {
            left++;
        }
        /* hledám zprava prvky menší než pivot */
        while (array[right] > pivot) {
            right--;
        }
        /* až najdu dva, vyměním je */
        if (left < right) {
            int tmp = array[left];
            array[left] = array[right];
            array[right] = tmp;
        }
        /* left a right se musí překřížit. 
         * Tj. left musí být vpravo a right vlevo*/
        if (left <= right) {
            left++;
            right--;
        }
    }
    /* Teď se left a right překříží, takže left je vpravo
     * a right je vlevo. */
    if (start < right) {
        quickSort(start, right, array);
    }
    if (left < stop) {
        quickSort(left, stop, array);
    }
}

/* *
 * Malé zrychlení algoritmu Quick sort, který je neefektivní 
 * na krátká pole. Ty se řadí pomocí Insert sort. 
 */
void myqSort(int start, int stop, int array[])  //%.h
{
    int pivot = array[(stop + start) / 2];
    int left = start;
    int right = stop;
    while (left < right) {
        /* hledám zleva prvky větší než pivot */
        while (array[left] < pivot) {
            left++;
        }
        /* hledám zprava prvky menší než pivot */
        while (array[right] > pivot) {
            right--;
        }
        /* až najdu dva, vyměním je */
        if (left < right) {
            int tmp = array[left];
            array[left] = array[right];
            array[right] = tmp;
        }
        /* left a right se musí překřížit. 
         * Tj. left musí být vpravo a right vlevo*/
        if (left <= right) {
            left++;
            right--;
        }
    }
    /* Teď se left a right překříží, takže left je vpravo
     * a right je vlevo. Krátká pole se řadí pomocí Insert sort, 
     * dlouhá pomocí Quick sort. */
    if (right - start > 22) {
        myqSort(start, right, array);
    } else {
        insertSort(start, right, array);
    }
    if (stop - left > 22) {
        myqSort(left, stop, array);
    } else {
        insertSort(left, stop, array);
    }
}

/**
 * Funkce vrací:
 * -1 == ( a<b )
 *  0 == ( a==b )
 *  1 == ( a>b ) 
 * Funkce je volána ve funkci sqSort
 */
int cmpAB(const void *a, const void *b)
{
    if (*((int *) a) < (*(int *) b)) {
        return -1;
    } else if (*((int *) a) == (*(int *) b)) {
        return 0;
/*    } else if (*((int *) a) > (*(int *) b)) {
        return 1;*/
    } else {
        return 1;
    }
}

/* 
 * Funkce pouze "obaluje" qsort z stdlib.h 
 */
void sqSort(int start, int stop, int array[])   //%.h
{
    qsort(&array[start], stop - start + 1, sizeof(array[start]), cmpAB);
}
